// Requires the use of MCUDude's Minicore addon for running the ATMEGA328 without XTAL
// and to access the two XTAL pins as digital IO

#include <LiquidCrystal.h>

int menu;
int select;
int steps;
int stepdelay;
int pieces;
float cutlength;

// these variables will calulate your length
// for material feeder spindle / wheel diameter 
float rolldiameter = 1.0;
// how many stepper motor steps to rotate feeder wheel one revolution 
// use test function to measure
float stepsperrotation = 200;

float lenperrotation;
float lenperstep;

const int PIN_LCD_RS = 8;
const int PIN_LCD_EN = 9;
const int PIN_LCD_D4 = 10;
const int PIN_LCD_D5 = 11;
const int PIN_LCD_D6 = 12;
const int PIN_LCD_D7 = 13;

const int PIN_STEPPER_0 = A0;
const int PIN_STEPPER_1 = A1;
const int PIN_STEPPER_2 = A2;
const int PIN_STEPPER_3 = A3;

const int PIN_BUTTON_UP = 6;
const int PIN_BUTTON_DOWN = 5;
const int PIN_BUTTON_LEFT = 4;
const int PIN_BUTTON_RIGHT = 3;
const int PIN_BUTTON_OK = 2;

const int PIN_SENSOR_1 = 7;
const int PIN_SENSOR_2 = 0;
const int PIN_SENSOR_3 = 1;

// requires minicore
const int PIN_SWITCH_1 = 20;
const int PIN_SWITCH_2 = 21;

const int PIN_ACTUATOR_1 = A4;
const int PIN_ACTUATOR_2 = A5;

// LCD dialog choices
const int MENU_MAIN = 1;
const int MENU_CUT_PIECES = 2;
const int MENU_CUT_LENGTH = 3;
const int MENU_CUT_CONFIRM = 4;
const int MENU_CUT = 5;
const int MENU_TEST = 6;
const int MENU_TEST_INPUT = 7;
const int MENU_TEST_OUTPUT = 8;



// initialize the library by associating any needed LCD interface pin
// with the arduino pin number it is connected to
LiquidCrystal lcd(PIN_LCD_RS, PIN_LCD_EN, PIN_LCD_D4, PIN_LCD_D5, PIN_LCD_D6, PIN_LCD_D7);

void setup() {
  // configure each pin for input / output
  // BUTTONs are all input with pullup enabled
  pinMode(PIN_BUTTON_UP, INPUT_PULLUP);
  pinMode(PIN_BUTTON_DOWN, INPUT_PULLUP);
  pinMode(PIN_BUTTON_LEFT, INPUT_PULLUP);
  pinMode(PIN_BUTTON_RIGHT, INPUT_PULLUP);
  pinMode(PIN_BUTTON_OK, INPUT_PULLUP);

  // optical SENSOR pin is input
  pinMode(PIN_SENSOR_1, INPUT);
  pinMode(PIN_SENSOR_2, INPUT);
  pinMode(PIN_SENSOR_2, INPUT);

  // SWITCH pin is input pullup enabled (these are XTAL on normal arduino boards)
  // this requires Minicore
  pinMode(PIN_SWITCH_1, INPUT_PULLUP);
  pinMode(PIN_SWITCH_2, INPUT_PULLUP);

  // STEPPER pins are output
  pinMode(PIN_STEPPER_0, OUTPUT);
  pinMode(PIN_STEPPER_1, OUTPUT);
  pinMode(PIN_STEPPER_2, OUTPUT);
  pinMode(PIN_STEPPER_3, OUTPUT);

  // ACTUATOR pins are also output 
  pinMode(PIN_ACTUATOR_1, OUTPUT);
  pinMode(PIN_ACTUATOR_2, OUTPUT);  
  
  // set up the LCD's number of columns and rows:
  lcd.begin(16, 2);

  lcd.clear();
  lcd.print("CUTBOT HF V1.2");
  delay(2000);

  steps = 0;
  stepdelay = 1;
  pieces = 0;
  cutlength= 0;
  menu = MENU_MAIN;
  select = 0;
  
}

void stepforward() { 
  digitalWrite(PIN_STEPPER_0, HIGH);
  delay(stepdelay);
  digitalWrite(PIN_STEPPER_0, LOW);
  delay(stepdelay);

  digitalWrite(PIN_STEPPER_2, HIGH);
  delay(stepdelay);
  digitalWrite(PIN_STEPPER_2, LOW);
  delay(stepdelay);

  digitalWrite(PIN_STEPPER_1, HIGH);
  delay(stepdelay);
  digitalWrite(PIN_STEPPER_1, LOW);
  delay(stepdelay);

  digitalWrite(PIN_STEPPER_3, HIGH);
  delay(stepdelay);
  digitalWrite(PIN_STEPPER_3, LOW);
  delay(stepdelay);
}

void stepback() {
  digitalWrite(PIN_STEPPER_3, HIGH);
  delay(stepdelay);
  digitalWrite(PIN_STEPPER_3, LOW);
  delay(stepdelay);

  digitalWrite(PIN_STEPPER_1, HIGH);
  delay(stepdelay);
  digitalWrite(PIN_STEPPER_1, LOW);
  delay(stepdelay);

  digitalWrite(PIN_STEPPER_2, HIGH);
  delay(stepdelay);
  digitalWrite(PIN_STEPPER_2, LOW);
  delay(stepdelay);

  digitalWrite(PIN_STEPPER_0, HIGH);
  delay(stepdelay);
  digitalWrite(PIN_STEPPER_0, LOW);
  delay(stepdelay);
}

void mainmenu() {
  select = MENU_CUT_PIECES;
  lcd.clear();
 
  lcd.setCursor(0, 0);
  lcd.print("MAIN FUNCTION");

  // selection logic
  while(true){
    lcd.setCursor(0, 1);
    if (select == MENU_CUT_PIECES){
      lcd.print(">CUT");
    }
    else lcd.print(" CUT");
  
    lcd.setCursor(5, 1);
    if (select == MENU_TEST){
      lcd.print(">TEST");
    }
    else lcd.print(" TEST");

    // button logic
    if (digitalRead(PIN_BUTTON_LEFT) == LOW){
      if (select == MENU_TEST) select = MENU_CUT_PIECES;
    }
    if (digitalRead(PIN_BUTTON_RIGHT) == LOW){
      if (select == MENU_CUT_PIECES) select = MENU_TEST;
    }
    if (digitalRead(PIN_BUTTON_OK) == LOW){
      menu = select;
      break;
    }
  }
}

void testmenu() {
  select = MENU_TEST_INPUT;
  lcd.clear();
 
  lcd.setCursor(0, 0);
  lcd.print("TEST FUNCTIONS");

  // selection logic
  while(true){
    lcd.setCursor(0, 1);
    if (select == MENU_TEST_INPUT){
      lcd.print(">INPUT");
    }
    else lcd.print(" INPUT");
  
    lcd.setCursor(7, 1);
    if (select == MENU_TEST_OUTPUT){
      lcd.print(">OUTPUT");
    }
    else lcd.print(" OUTPUT");

    // button logic
    if (digitalRead(PIN_BUTTON_LEFT) == LOW){
      if (select == MENU_TEST_OUTPUT) select = MENU_TEST_INPUT;
    }
    if (digitalRead(PIN_BUTTON_RIGHT)== LOW){
      if (select == MENU_TEST_INPUT) select = MENU_TEST_OUTPUT;
    }
    if (digitalRead(PIN_BUTTON_OK) == LOW){
      menu = select;
      break;
    }
    if (digitalRead(PIN_BUTTON_UP) == LOW){
      menu = MENU_MAIN;
      break;
    }
  }
}

void getcutlength() {
  lcd.clear();
 
  // set the cursor to column 0, line 1
  lcd.setCursor(0, 0);
  lcd.print("How long?");

  while(true){
    lcd.setCursor(0, 1);
    lcd.print(cutlength);
    delay(100);
  
    if (digitalRead(PIN_BUTTON_UP) == LOW) cutlength+=.1;
    if (digitalRead(PIN_BUTTON_DOWN) == LOW) cutlength-=.1;
    if (digitalRead(PIN_BUTTON_RIGHT) == LOW) cutlength+=1;  
    if (digitalRead(PIN_BUTTON_LEFT) == LOW) cutlength-=1;
    if (digitalRead(PIN_BUTTON_OK) == LOW){
      menu = MENU_CUT_CONFIRM;
      break;
    }
  }
}

void getcutpieces() {
  lcd.clear();

  // set the cursor to column 0, line 1
  lcd.setCursor(0, 0);
  lcd.print("How many pieces?");

  while(true){
    lcd.setCursor(0, 1);
    lcd.print(pieces);
    delay(100);
  
    if (digitalRead(PIN_BUTTON_UP) == LOW) pieces++;
    if (digitalRead(PIN_BUTTON_DOWN) == LOW) pieces--;
    if (digitalRead(PIN_BUTTON_RIGHT) == LOW) pieces+=10;
    if (digitalRead(PIN_BUTTON_LEFT) == LOW) pieces-=10;
    if (digitalRead(PIN_BUTTON_OK) == LOW){
      menu = MENU_CUT_LENGTH;
      break;
    }
  }
}

void cutconfirm() {
  select = MENU_MAIN;
  lcd.clear();

  // set the cursor to column 0, line 1
  lcd.setCursor(0, 0);
  lcd.print("Cut ");
  lcd.print(pieces);
  lcd.print(" x ");
  lcd.print(cutlength);
  lcd.print("?");

  // selection logic
  while(true){
    // next line
    lcd.setCursor(0, 1);
    if (select == MENU_MAIN){
      lcd.print(">NO");
    }
    else lcd.print(" NO");
  
    lcd.setCursor(5, 1);
    if (select == MENU_CUT){
      lcd.print(">YES");
    }
    else lcd.print(" YES");

    // button logic
    if (digitalRead(PIN_BUTTON_LEFT) == LOW){
      if (select == MENU_CUT) select = MENU_MAIN;
    }
    if (digitalRead(PIN_BUTTON_RIGHT) == LOW){
      if (select == MENU_MAIN) select = MENU_CUT;
    }
    if (digitalRead(PIN_BUTTON_OK) == LOW){
      menu = select;
      break;
    }
  }
}

void docut() {
  int i;
  int j;
  int attempt;
  int maxcuts = 3;

  // length per rotation is distance travelled in one complete rotation of the stepper 
  lenperrotation = rolldiameter * 3.14159;

  // length per step is the distance covered in on step of the stepper 
  lenperstep = lenperrotation / stepsperrotation;

  // steps is what is needed to get the right length
  steps = cutlength / lenperstep;

  // set the cursor to column 0, line 1
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Resetting...");

  // position the cutter for start
  digitalWrite(PIN_ACTUATOR_1, HIGH);
  while (digitalRead(PIN_SENSOR_1) == 1) delay(10);
  while (digitalRead(PIN_SENSOR_1) == 0) delay(10);
  digitalWrite(PIN_ACTUATOR_1, LOW);

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Cutting...");
  lcd.setCursor(0, 1);
  lcd.print("Press OK to stop");

  // cutting logic
  for(i = 0; i < pieces; i++){
    
    // make sure material is available
    while (digitalRead(PIN_SENSOR_2) == 0){
      if (digitalRead(PIN_BUTTON_OK) == LOW){
        menu = MENU_MAIN;
        return;
      }
      lcd.setCursor(0, 0);
      lcd.print("Nothing to cut!");
      lcd.setCursor(0, 1);
      lcd.print("Press OK to stop");
      delay(100);
      
    }
    
    // print progress message
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Cutting ");
    lcd.print(i);
    lcd.print(" / ");
    lcd.print(pieces);
    
    //we move forward the right length that's required
    for(j = 0; j < steps; j++){
      stepforward();
    }
    
    // check to make sure there is something there to cut (at end)
    if (digitalRead(PIN_SENSOR_3) == 0){
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Possible Jam!");
      lcd.setCursor(0, 1);
      lcd.print("OK to continue");
      while (digitalRead(PIN_BUTTON_OK) == 1) delay(10);
    }

    // try cutting multiple times if inital cut fails
    attempt = 0;
    while (attempt < maxcuts){ 

      // do the cut, check for abort button press
      attempt++;
      digitalWrite(PIN_ACTUATOR_1, HIGH);
      while (digitalRead(PIN_SENSOR_1) == 1){
        if (digitalRead(PIN_BUTTON_OK) == LOW){
          digitalWrite(PIN_ACTUATOR_1, LOW);
          menu = MENU_MAIN;
          break;
          delay(10);
        }
      }
      while (digitalRead(PIN_SENSOR_1) == 0){
        // button logic - abort
        if (digitalRead(PIN_BUTTON_OK) == LOW){
          digitalWrite(PIN_ACTUATOR_1, LOW);
          menu = MENU_MAIN;
          break;
          delay(10);
        }
      }
      
      // check to make sure it's cut
      if (digitalRead(PIN_SENSOR_3) == 1){
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("Retrying");
        continue;
      }
      else break;
    }
    digitalWrite(PIN_ACTUATOR_1, LOW);
    // if max attempts reached
    if (attempt == maxcuts){
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Possible Jam!");
      lcd.setCursor(0, 1);
      lcd.print("OK to continue");
      while (digitalRead(PIN_BUTTON_OK) == 1) delay(10);
    }

    // button logic - quit
    if (digitalRead(PIN_BUTTON_OK) == LOW){
      menu = MENU_MAIN;
      break;
    }
  }
  menu = MENU_MAIN;
}

  
void testinput() {
  int okbtn = HIGH;
  lcd.clear();
  
  // set the cursor to column 0, line 1
  lcd.setCursor(0, 0);
  lcd.print("INPUT TEST");

  while(true){
    if (okbtn == LOW && digitalRead(PIN_BUTTON_OK) == HIGH){
      menu = MENU_TEST;
      break;
    }
    okbtn = digitalRead(PIN_BUTTON_OK);
    lcd.setCursor(0, 1);
    lcd.print("B");
    lcd.print(digitalRead(PIN_BUTTON_UP));
    lcd.print(digitalRead(PIN_BUTTON_DOWN));
    lcd.print(digitalRead(PIN_BUTTON_LEFT));
    lcd.print(digitalRead(PIN_BUTTON_RIGHT));
    lcd.print(digitalRead(PIN_BUTTON_OK));
    lcd.print(" O");
    lcd.print(digitalRead(PIN_SENSOR_1));
    lcd.print(digitalRead(PIN_SENSOR_2));
    lcd.print(digitalRead(PIN_SENSOR_3));
    lcd.print(" S");
    lcd.print(digitalRead(PIN_SWITCH_1));
    lcd.print(digitalRead(PIN_SWITCH_2));
  }
}

void testoutput() {
  lcd.clear();

  // set the cursor to column 0, line 1
  lcd.setCursor(0, 0);
  lcd.print("OUTPUT TEST");

  while (true){
    lcd.setCursor(0, 1);

    // steps, format so it fits on the screen
    if (steps >= 0){
      lcd.print("F");
      if (steps < 10) lcd.print("0");
      if (steps < 100) lcd.print("0");
      lcd.print(steps % 1000);
    }
    else{
      lcd.print("R");
      if (steps > -10) lcd.print("0");
      if (steps > -100) lcd.print("0");
      lcd.print(-steps % 1000);
    }

    // actuators
    lcd.print(" A");
    lcd.print(digitalRead(PIN_ACTUATOR_1));
    lcd.print(digitalRead(PIN_ACTUATOR_2));
    // optical sensors
    lcd.print(" O");
    lcd.print(digitalRead(PIN_SENSOR_1));
    lcd.print(digitalRead(PIN_SENSOR_2));
    lcd.print(digitalRead(PIN_SENSOR_3));
    // switches
    lcd.print(" S");
    lcd.print(digitalRead(PIN_SWITCH_1));
    lcd.print(digitalRead(PIN_SWITCH_2));
    lcd.print("   ");
    
    // test stepper here
    if (digitalRead(PIN_BUTTON_RIGHT) == LOW) {
      stepforward(); 
      steps=steps+1;
    }
    else if (digitalRead(PIN_BUTTON_LEFT) == LOW) {
      stepback(); 
      steps=steps-1;
    }

    // test cutter here
    if (digitalRead(PIN_BUTTON_UP) == LOW) {
      digitalWrite(PIN_ACTUATOR_1, HIGH);
    }
    else digitalWrite(PIN_ACTUATOR_1, LOW);

    if (digitalRead(PIN_BUTTON_DOWN) == LOW) {
      digitalWrite(PIN_ACTUATOR_2, HIGH);
    }
    else digitalWrite(PIN_ACTUATOR_2, LOW);

    // exit if ok pressed
    if (digitalRead(PIN_BUTTON_OK) == LOW){
      menu = MENU_TEST;
      break;
    }
  }
}


void loop(){
  if (menu == MENU_MAIN) mainmenu();

  // cut screens
  else if (menu == MENU_CUT_PIECES) getcutpieces();
  else if (menu == MENU_CUT_LENGTH) getcutlength();
  else if (menu == MENU_CUT_CONFIRM) cutconfirm();
  else if (menu == MENU_CUT) docut();
  
  
  // test functions
  else if (menu == MENU_TEST) testmenu();
  else if (menu == MENU_TEST_OUTPUT) testoutput();
  else if (menu == MENU_TEST_INPUT) testinput();

  // make sure ok button released between screens
  while (digitalRead(PIN_BUTTON_OK) == LOW) delay(50);
  
  //testinput();  
}
